#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H  */

#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/config_api.h>
#include <net-snmp/output_api.h>
#include <net-snmp/mib_api.h>

#include "docsis.h"
#include "docsis_globals.h"
#include "ethermac.h"
#include "md5.h"

int _total_children = 0;

void bury_zombie(int);
void handle_request(char*);
void handle_conn(int);
int server_init(int, char*);

void bury_zombie(int s)
{
  pid_t pid;
  int status;
                
  while((pid = wait3(&status, WNOHANG, NULL)) > 0);
  __sync_fetch_and_add(&_total_children, -1);
}

                                

/* server_init - initialize the server */
int  server_init(int port, char * address)
{
	struct sockaddr_in sin;
	int listeningSocket;
	int on = 1;
	
	/* Acquire socket */
	if((listeningSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
	{
		perror("socket()");
		return(-1);
	}
	/* kill "Address already in use" error message */
	if (setsockopt(listeningSocket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int)) == -1) 
	{
		perror("setsockopt(SO_REUSEADDR)");
		exit(1);
	}
	/* Bind it to a port */
	sin.sin_family = PF_INET;
	sin.sin_port = htons(port);
	if(address)
	{
		if(!inet_aton(address, &sin.sin_addr))
		{
			perror("inet_aton()");
			return(-1);
		}
	}
	else
		sin.sin_addr.s_addr = INADDR_ANY;
	
	if(bind(listeningSocket, (struct sockaddr*) &sin, sizeof(sin)) == -1)
	{
		perror("bind()");
		return(-1);
	}
	
	/* Configure listening */
	if(listen(listeningSocket, 10) == -1)
	{
		perror("listen()");
		return(-1);
	}

	/* Setup non-blocking operation */
	#if 0
	if(fcntl(listeningSocket, F_SETFL, O_NONBLOCK) == -1)
	{
		perror("fcntl(listeningSocket, F_SETFL, O_NONBLOCK)");
		return(-1);
	}
	#endif
	
	return(listeningSocket);
}

void handle_request(char *buf)
{
#define ARGV_SIZE 64
  char *argv[ARGV_SIZE + 1];
  int argc = 0;
  
  /* 1 - replace \r by zeros */
  while(argc < ARGV_SIZE)
  {
    char * i = index(buf, '\r');
    if(!i) i = index(buf, '\n');
    if(!i) break;
    
    while(*i == '\r' || *i == '\n')
      *(i++) = 0;

    if(strlen(buf))
      argv[argc++] = buf;
    
    buf = i;
  }
  
  main_instance(argc, argv);
  
}


void handle_conn(int conn)
{
#define ARGBUF_SIZE 1024
  char argbuf[ARGBUF_SIZE + 1];
  int l = 0;
  
  while(l < ARGBUF_SIZE)
  {
    int t;
    
    if((t = read(conn, argbuf + l, ARGBUF_SIZE - l)) <= 0)
    {
      perror("read");
      break;
    }
    
    l += t;
    argbuf[l] = 0;
    if(strstr(argbuf, "\n\n") || strstr(argbuf, "\r\n\r\n"))
    {
      /* We've got a packet */
      handle_request(argbuf);
      return;
    }
  }
  return;
}


int main(int argc, char **argv)
{
  int c, socket, conn;
  int port = 5560, timeout = 10, children_limit = 5;
  char *address = "127.0.0.1";
  
  while((c = getopt(argc, argv, "t:p:a:hl:")) != EOF)
  {
    switch(c)
    {
      case 'a':
        assert(address = strdup(optarg));
        break;
      case 'p':
        assert((port = atoi(optarg)) != -1);
        break;
      case 't':
        assert((timeout = atoi(optarg)) != -1);
        break;
      case 'l':
        assert((children_limit = atoi(optarg)) != -1);
        break;
      case 'h': case '?':
        fprintf(stderr, "Usage: %s [-a <IP address>] [-p <port>] [-t <compilation timeout>] [-l <child processes limit>]\n", argv[0]);
        return 1;
    }
  }
  main_init();
  
  if((socket = server_init(port, address)) == -1)
  {
    fprintf(stderr, "Cannot initialize TCP server\n");
    return 1;
  }
  
  signal(SIGCHLD, bury_zombie);
  
  while((conn = accept(socket, NULL, NULL)) != -1)
  {
    while(_total_children >= children_limit)
    {
      /* wait */
      usleep(50000);
    }
    
    fprintf(stderr, "fork(), total children = %d\n", _total_children);    
    switch(fork())
    {
      case -1:
        perror("fork");
        break;
      case 0:
        if(timeout) alarm(timeout);
        dup2(conn, 0);
        dup2(conn, 1);
        dup2(conn, 2);
        handle_conn(conn);
        return 0;
      default:
        __sync_fetch_and_add(&_total_children, 1);
        close(conn);
        break;
    }
      
  }
  
  perror("accept");

  return 0;  
}

